import asyncio
import concurrent.futures
import functools
import threading

thread_local = threading.local()
default_executor = concurrent.futures.ThreadPoolExecutor(
    thread_name_prefix=f"{__package__}:executor"
)


def run_in_executor(func=None, *, context=None, executor=None):
    """
    Decorator to run synchronous function in the given or default executor

    Optionally within the given async context manager (e.g. to limit concurrency
    with `asyncio.Lock` or `asyncio.Semaphore`)
    """
    if executor is None:
        executor = default_executor

    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            loop = asyncio.get_running_loop()

            # Use dummy Lock() as contextlib.nullcontext for older python versions
            current_context = context or asyncio.Lock()

            stop_event = threading.Event()

            def entrypoint():
                thread_local.stop_event = stop_event
                return func(*args, **kwargs)

            async with current_context:
                try:
                    return await loop.run_in_executor(executor, entrypoint)
                finally:
                    stop_event.set()

        return wrapper

    return decorator(func) if func else decorator


def maybe_cancel():
    """
    Checks if the current function (running with `run_in_executor`) was cancelled.
    If so, raises `asyncio.CancelledError` and stop execution.
    The exception will be ignored, so there's no need to catch it.
    """
    stop_event = getattr(thread_local, "stop_event", None)

    if stop_event and stop_event.is_set():
        raise asyncio.CancelledError


run_in_executor_locked = functools.partial(run_in_executor, context=asyncio.Lock())



import inspect

from delphai_utils.keycloak import decode_token
from functools import wraps
from grpc import StatusCode
from grpc.aio import ServicerContext
from typing import Dict, List, Tuple, Union


def get_authorization_token(context: ServicerContext) -> str:
    """
    Get user access token.
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :return access token.
      More information: https://wiki.delphai.dev/wiki/Authorization
    :rtype: str
    """

    metadata = dict(context.invocation_metadata())
    if "authorization" not in metadata:
        return None

    authorization_header = metadata["authorization"]
    access_token = authorization_header.split("Bearer ")[1]

    return access_token


async def get_user(context: ServicerContext) -> Dict:
    """
    Get x-delphai-user information.
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :raises: KeyError
    :return x-delphai-user dictionary.
      More information: https://wiki.delphai.dev/wiki/Authorization
    :rtype: Dict
    """

    access_token = get_authorization_token(context)
    if not access_token:
        return {}
    decoded_access_token = await decode_token(access_token)
    user = {
        "https://delphai.com/mongo_user_id": decoded_access_token.get("mongo_user_id"),
        "https://delphai.com/client_id": decoded_access_token.get("mongo_client_id"),
        "user_id": decoded_access_token.get("sub"),
    }
    if "realm_access" in decoded_access_token and "roles" in decoded_access_token.get(
        "realm_access"
    ):
        roles = decoded_access_token.get("realm_access").get("roles")
        user["roles"] = roles
    if "group_membership" in decoded_access_token:
        user["groups"] = decoded_access_token["group_membership"]
    if "limited_dataset_group_name" in decoded_access_token:
        user["limited_dataset_group_name"] = decoded_access_token[
            "limited_dataset_group_name"
        ]
    if "name" in decoded_access_token:
        user["name"] = decoded_access_token["name"]
    if decoded_access_token.get("groups"):
        user_groups = decoded_access_token["groups"][0]
        user["customer_id"] = user_groups[0]["id"]
        if len(user_groups) > 1:
            user["department_id"] = user_groups[-1]["id"]
    if decoded_access_token.get("preferred_currency"):
        user["preferred_currency"] = decoded_access_token["preferred_currency"]

    return user


async def get_user_and_client_ids(context: ServicerContext) -> Tuple[str, str]:
    """
    Get user_id and client_id.
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :return user_id and client_id.
      More information: https://wiki.delphai.dev/wiki/Authorization
    :rtype: Tuple[str, str]
    """
    try:
        user = await get_user(context)
        user_id = user.get("https://delphai.com/mongo_user_id")
        client_id = user.get("https://delphai.com/client_id")
    except Exception:
        return "", ""
    return user_id, client_id


async def get_roles(context: ServicerContext):
    """
    Utility function to retrieve roles from grpc context.
    :param Union[List[str], grpc.aio.ServicerContext] context:
      a list of roles or the grpc context passed to class methods
    :return list of roles
    :rtype List[str]
    """

    try:
        user = await get_user(context)
        return user.get("roles") or []
    except KeyError:
        return []


async def get_groups(context: ServicerContext) -> List[str]:
    """
    Gets groups of calling identity
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :return raw roles passed from keycloak.
      For example this can be ["/delphai/Development"]
    :rtype List[str]
    """

    assert isinstance(context, object)
    try:
        user = await get_user(context)
    except KeyError:
        return []
    return user.get("groups") or []


async def get_affiliation(
    context: ServicerContext,
) -> Tuple[Union[str, None], Union[str, None]]:
    """
    Gets organization and department of user
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :return organization and department as a tuple or None if not affiliated
    :rtype Union[Tuple[str, None], Tuple[str, None]]
    """

    raw_groups = await get_groups(context)
    if len(raw_groups) == 0:
        return None, None
    else:
        group_hirarchy = raw_groups[0].split("/")[1:3]
        if len(group_hirarchy) == 1:
            return group_hirarchy.pop(), None
        elif len(group_hirarchy) == 2:
            return tuple(group_hirarchy)
        else:
            return None, None


async def is_authorized(
    provided_roles: Union[List[str], ServicerContext],
    required_roles: List[str],
    decision_logic: str = "all",
    abort: bool = False,
):
    """
    Makes a decision if user is authorized based on intersection of roles.
    :param Union[List[str], grpc.aio.ServicerContext] provided_roles:
      a list of roles or the grpc context passed to class methods
    :param List[str] required_roles: roles the request must pass
    :param str decision_logic: one of 'all' or 'some'. Default: 'all'
    :param bool abort: if set to true grpc call will abort with PERMISSION_DENIED. Default: False
    :rtype: bool
    """

    assert len(required_roles) != 0
    assert type(required_roles) == list
    assert any([allowed == decision_logic for allowed in ["all", "some"]])
    resolved_roles = await get_roles(provided_roles)

    required_roles = set(required_roles)  # in case same role is passed mulitple times
    role_intersection = required_roles.intersection(resolved_roles)
    all_intersects = len(role_intersection) == len(required_roles)
    some_intersects = len(role_intersection) > 0
    _is_authorized = (decision_logic == "all" and all_intersects) or (
        decision_logic == "some" and some_intersects
    )
    is_grpc_context = type(provided_roles) is not list
    if abort and not _is_authorized and is_grpc_context:
        details = f"Provided roles {resolved_roles} don't intersect with required roles {required_roles}."
        provided_roles.abort(StatusCode.PERMISSION_DENIED, details)
    return _is_authorized


def authorize(required_roles: List[str], decision_logic: str = "all"):
    """
    Decorator for grpc endpoints to restrict role-based access.
    :param List[str] required_roles: list of strings which roles are required
    :param str decision_logic: one of 'all' or 'some'. Default: 'all'
    :return: wraped function
    :rtype: func
    """

    def wrap(func):
        assert inspect.iscoroutinefunction(func)

        @wraps(func)
        async def wrapped_func(self, request, context, *args, **kwargs):
            if await is_authorized(context, required_roles, decision_logic):
                return await func(self, request, context, *args, **kwargs)
            else:
                details = f"Given roles don't intersect with {required_roles} using decission logic '{decision_logic}'"
                await context.abort(StatusCode.PERMISSION_DENIED, details)

        return wrapped_func

    return wrap



from uuid import uuid5, NAMESPACE_URL
from datetime import datetime, timezone


def clean_url(url, keep_www=False):
    """
    Format and clean an url to be saved or checked.
    Args:
        url: url to be formatted
        keep_www: keep the 'www' part of the url
    Returns: formatted url
    """

    url = url.strip()
    url = url.replace("https://", "").replace("http://", "").rstrip("/")
    if not keep_www:
        url = url.replace("www.", "")
    split_url = url.split("/")
    split_url[0] = split_url[0].lower()
    return "/".join(split_url)


def get_clean_domain(url):
    """
    Format and clean an url and returns domain.
    Args:
        url: url to be formatted
    Returns: formatted domain
    """

    return clean_url(url).split("/")[0]


def create_blob_reference(url: str, timestamp: datetime, blob_type: str):
    """
    Creates a a blob reference using the domain of the URL, a UUID5 of the URL, the timestamps ISO format and the given blob type.
    Args:
        url: Url representing the blob
        timestamp: The timezone-aware timestamp of the blob
        blob_type: The type of the blob. Like "html" or "content"
    """

    if not timestamp.tzinfo:
        raise ValueError("timestamp must be timezone-aware")
    timestamp = timestamp.astimezone(timezone.utc)

    return f"{get_clean_domain(url)}/{uuid5(NAMESPACE_URL, url)}/{timestamp.isoformat()}/{blob_type}"



from httpx import (
    StreamError,
    TimeoutException,
    NetworkError,
    ProtocolError,
    ProxyError,
    HTTPStatusError,
)
from tenacity import retry as tenacity_retry
from tenacity import (
    retry_if_exception_type,
    retry_if_exception,
    stop_after_attempt,
    wait_random_exponential,
)

HTTP_RETRY_EXCEPTION_TYPES = (
    TimeoutException,
    NetworkError,
    ProtocolError,
    ProxyError,
    StreamError,
)


def test_response_status_code(exception):
    if not isinstance(exception, HTTPStatusError):
        return False

    if 500 <= exception.response.status_code < 600:
        return True

    return exception.response.status_code in (
        408,  # Request Timeout
        429,  # Too Many Requests
    )


def retry_httpx(
    stop=stop_after_attempt(3),
    wait=wait_random_exponential(multiplier=1, max=60, min=1),
    retry=retry_if_exception_type(tuple(HTTP_RETRY_EXCEPTION_TYPES))
    | retry_if_exception(test_response_status_code),
    reraise=True,
    **kwargs,
):
    return tenacity_retry(stop=stop, wait=wait, retry=retry, reraise=reraise, **kwargs)



import asyncio
import hypercorn
import json
import logging
import os

from typing import Optional
from urllib.parse import urlparse

from google.protobuf import symbol_database
from google.protobuf.json_format import MessageToDict
from grpc import Server, StatusCode
from grpc.experimental.aio import AioRpcError, insecure_channel
from hypercorn.asyncio import serve as hypercorn_serve
from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse

from delphai_utils.utils import find_free_port, generate_default_message

logger = logging.getLogger(__name__)

SUPPORTED_METHODS = ["get", "put", "post", "delete", "patch"]
STATUS_MAP = {
    StatusCode.OK: 200,
    StatusCode.CANCELLED: 499,
    StatusCode.UNKNOWN: 500,
    StatusCode.INVALID_ARGUMENT: 400,
    StatusCode.DEADLINE_EXCEEDED: 504,
    StatusCode.NOT_FOUND: 404,
    StatusCode.ALREADY_EXISTS: 409,
    StatusCode.PERMISSION_DENIED: 403,
    StatusCode.UNAUTHENTICATED: 401,
    StatusCode.RESOURCE_EXHAUSTED: 429,
    StatusCode.FAILED_PRECONDITION: 412,
    StatusCode.ABORTED: 499,
    StatusCode.OUT_OF_RANGE: 416,
    StatusCode.UNIMPLEMENTED: 501,
    StatusCode.INTERNAL: 500,
    StatusCode.UNAVAILABLE: 503,
    StatusCode.DATA_LOSS: 420,
}
status_map = STATUS_MAP  # for backward compatibility, don't use it


async def http_exception_handler(request, exc):
    if "/favicon.ico" in request.url.path:
        detail = "Not Found"
        status_code = 404
    else:
        path = urlparse(str(request.url)).path
        logger.error(f"[{exc.status_code}] {request.method} {path} - {exc.detail}")
        detail = exc.detail
        status_code = exc.status_code
    return JSONResponse(
        {"detail": detail, "status": exc.status_code}, status_code=status_code
    )


def _add_method_handlers(method_descriptor, app, channel):
    logger.info(f"  processing {method_descriptor.name}")

    async def method_get_handler(request: Request):
        return JSONResponse(
            {
                "function_name": method_descriptor.full_name,
                "input": generate_default_message(method_descriptor.input_type),
                "output": generate_default_message(method_descriptor.output_type),
            }
        )

    async def method_post_handler(request: Request):
        metadata = request.headers.items()

        body = await request.body()
        body_params = json.loads(body) if body else {}

        input = {**request.path_params, **request.query_params, **body_params}

        try:
            input_prototype = symbol_database.Default().GetPrototype(
                method_descriptor.input_type
            )
            output_prototype = symbol_database.Default().GetPrototype(
                method_descriptor.output_type
            )
            method_callable = channel.unary_unary(
                f"/{method_descriptor.containing_service.full_name}/{method_descriptor.name}",
                request_serializer=input_prototype.SerializeToString,
                response_deserializer=output_prototype.FromString,
            )
            response = await method_callable(
                input_prototype(**input), metadata=metadata
            )

            output = MessageToDict(
                response,
                preserving_proto_field_name=True,
                including_default_value_fields=True,
            )
            return JSONResponse(output)

        except AioRpcError as ex:
            detail = ex.details()
            grpc_status = ex.code()
            http_status_code = STATUS_MAP[grpc_status]
            raise HTTPException(http_status_code, detail=detail)

        except Exception as ex:
            detail = str(ex).replace("\n", " ")
            http_status_code = 500
            raise HTTPException(http_status_code, detail=detail)

    app.add_route(
        f"/{method_descriptor.full_name}",
        route=method_get_handler,
        methods=["get"],
    )
    app.add_route(
        f"/{method_descriptor.full_name}",
        route=method_post_handler,
        methods=["post"],
    )

    for field_descriptor, http_rule in method_descriptor.GetOptions().ListFields():
        if field_descriptor.full_name != "google.api.http":
            continue

        for method in SUPPORTED_METHODS:
            http_path = getattr(http_rule, method)
            if http_path:
                app.add_route(
                    http_path,
                    route=method_post_handler,
                    methods=[method],
                )


def start_gateway(
    server: Server,
    grpc_port: int,
    http_port: Optional[int] = None,
    app: Optional[Starlette] = None,  # main application to add GRPC routes to it
    public_api: Optional[Starlette] = None,
):
    if not app:
        debug = os.environ.get("DELPHAI_ENVIRONMENT") == "development"
        app = Starlette(debug=debug)

        # This module registers its metrics on import
        # Import it only if we need it
        from starlette_prometheus import PrometheusMiddleware

        app.add_middleware(PrometheusMiddleware)

    app.add_exception_handler(HTTPException, http_exception_handler)
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_methods=["*"],
        allow_headers=["*"],
        allow_credentials=True,
    )

    from .grpc_server import GRPC_OPTIONS

    channel = insecure_channel(f"localhost:{grpc_port}", options=GRPC_OPTIONS)

    if not hasattr(server, "descriptors"):
        raise RuntimeError(
            "Server instance does not include the proto file descriptors. Make sure you instantiate it with 'create_grpc_server'"
        )
    for file_descriptor in server.descriptors:
        for service_descriptor in file_descriptor.services_by_name.values():
            if service_descriptor.full_name.startswith("grpc."):
                logger.info(f"skipping service {service_descriptor.name}")
                continue

            logger.info(f"processing service {service_descriptor.name}")
            for method_descriptor in service_descriptor.methods_by_name.values():
                logger.info(f"  processing {method_descriptor.name}")
                _add_method_handlers(method_descriptor, app, channel)

    if not http_port:
        http_port = find_free_port(7070)

    logger.info(f"starting gateway on port {http_port}")
    config = hypercorn.Config()
    config.bind = [f"0.0.0.0:{int(http_port)}"]
    config.accesslog = logger
    config.access_log_format = '[%(s)s] %(m)s %(U)s?%(q)s [%(L)ss] %(b)s "%(h)s"'
    config.errorlog = logger

    if public_api:
        app = app.with_public_api(public_api)

    return asyncio.get_event_loop().create_task(hypercorn_serve(app, config))
